-- Rider Bump
-- removes mid-air playerbumps and nerfs ground-based ones
-- you're free to edit and use this code for whatever; most of this is already taken from the game's source

local bumptime = (TICRATE*100)/583 
local wipeoutslowtime = 20

local function K_GetMobjWeight(mobj, against, nerfVal)
	local weight = 5<<FRACBITS

	if mobj.type == MT_PLAYER then
		if not mobj.player then return end

		if against.player and not against.player.kartstuff[k_spinouttimer] and mobj.player.kartstuff[k_spinouttimer] then
			weight = 0 -- Do not bump
		else
			weight = mobj.player.kartweight<<FRACBITS
			if mobj.player.speed > K_GetKartSpeed(mobj.player) then
				weight = $ + (mobj.player.speed - K_GetKartSpeed(mobj.player))/8
			end
		end
	elseif mobj.type == MT_FALLINGROCK then
		if against.player then
			if against.player.kartstuff[k_invincibilitytimer] or against.player.kartstuff[k_growshrinktimer] > 0 then
				weight = 0
			else
				weight = against.player.kartweight<<FRACBITS
			end
		end
	elseif mobj.type == MT_ORBINAUT or mobj.type == MT_ORBINAUT_SHIELD then -- orbis
		if against.player then
			weight = against.player.kartweight<<FRACBITS
		end
	elseif mobj.type == MT_JAWZ or mobj.type == MT_JAWZ_DUD or mobj.type == MT_JAWZ_SHIELD then
		if against.player then
			weight = (against.player.kartweight+3)<<FRACBITS
		else
			weight = 8<<FRACBITS
		end

	end

	if nerfVal then weight = $ / abs(nerfVal) end
	return weight
end


local function KP_KartBouncing(mobj1, mobj2, bounce, solid, nerf)
	local p1, p2 = mobj1.player, mobj2.player

	-- Don't bump when you're being reborn
	if (p1 and p1.playerstate ~= PST_LIVE)
	or (p2 and p2.playerstate ~= PST_LIVE) then
		return
	end

	if (p1 and p1.kartstuff[k_respawn])
	or (p2 and p2.kartstuff[k_respawn]) then
		return
	end

	-- Don't bump if you're flashing
	local flash

	if p1 then flash = K_GetKartFlashing(p1) end
	if p1 and p1.powers[pw_flashing] > 0 and p1.powers[pw_flashing] < flash then
		if p1.powers[pw_flashing] < flash-1 then
			p1.powers[pw_flashing] = $ + 1
		end
		return
	end


	if p2 then flash = K_GetKartFlashing(p2) end -- WHY IS THERE A NILCHECK HERE
	if p2 and p2.powers[pw_flashing] > 0 and p2.powers[pw_flashing] < flash then
		if (p2.powers[pw_flashing] < flash-1)
			p2.powers[pw_flashing] = ($ + 1)
		end
		return
	end

	-- Don't bump if you've recently bumped
	if p1 and p1.phs_justbumped then
		p1.phs_justbumped = bumptime
		return
	end

	if p2 and p2.phs_justbumped then
		p2.phs_justbumped = bumptime
		return
	end

	local mass1 = K_GetMobjWeight(mobj1, mobj2, 1)
	local mass2

	if solid and mass1 > 0 then
		mass2 = mass1
	else
		mass2 = K_GetMobjWeight(mobj2, mobj1, 1)
	end

	local momdifx = mobj1.momx - mobj2.momx
	local momdify = mobj1.momy - mobj2.momy

	-- Adds the OTHER player's momentum times a bunch, for the best chance of getting the correct direction
	local distx = (mobj1.x + mobj2.momx*3) - (mobj2.x + mobj1.momx*3)
	local disty = (mobj1.y + mobj2.momy*3) - (mobj2.y + mobj1.momy*3)

	if distx == 0 and disty == 0 then return end -- if there's no distance between the 2, they're directly on top of each other, don't run this

	-- Normalize distance to the sum of the two objects' radii, since in a perfect world that would be the distance at the point of collision...
	local dist = P_AproxDistance(distx, disty) or 1
	local nx = FixedDiv(distx, dist)
	local ny = FixedDiv(disty, dist)
	
	distx = FixedMul(mobj1.radius + mobj2.radius, nx)
	disty = FixedMul(mobj1.radius + mobj2.radius, ny)

	if momdifx == 0 or momdify == 0 then
		-- If there's no momentum difference, they're moving at exactly the same rate. Pretend they moved into each other.
		momdifx = -nx
		momdify = -ny
	end

	-- if the speed difference is less than this let's assume they're going proportionately faster from each other
	if P_AproxDistance(momdifx, momdify) < (25*mapobjectscale) then
		local momdiflength = P_AproxDistance(momdifx, momdify)
		local normalisedx = FixedDiv(momdifx, momdiflength)
		local normalisedy = FixedDiv(momdify, momdiflength)
		momdifx = FixedMul(25*mapobjectscale, normalisedx)
		momdify = FixedMul(25*mapobjectscale, normalisedy)
	end

	local dot = FixedMul(momdifx, distx) + FixedMul(momdify, disty)

	if dot >= 0 then return end -- They're moving away from each other

	local force = FixedDiv(dot, FixedMul(distx, distx)+FixedMul(disty, disty))

	if bounce or mass2 > 0 then -- Perform a Goomba Bounce.
		mobj1.momz = -mobj1.momz
	else
		local newz = mobj1.momz
		if mass2 > 0 then mobj1.momz = mobj2.momz end
		if mass1 > 0 and solid == false then mobj2.momz = newz end
	end
	
	local a = mobj1.angle
	local b = mobj2.angle - a
	
	local bonp = false
	
	//print((abs(b)/ANG1))
	
	if abs(b)/ANG1 <= 120 then
		bonp = true
	end	
	
	--if bonp then print("bonnnp") end

	if mass2 > 0 then
		if not bonp then
			mobj1.momx = mobj1.momx - (FixedMul(FixedMul(FixedDiv(2*mass2, mass1 + mass2), force), distx)/nerf)
			mobj1.momy = mobj1.momy - (FixedMul(FixedMul(FixedDiv(2*mass2, mass1 + mass2), force), disty)/nerf)
		else -- P1 gets the bump instead, don't fuck with P2
			mobj2.momx = mobj2.momx - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -distx)/nerf)
			mobj2.momy = mobj2.momy - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -disty)/nerf)
		end
	end

	if mass1 > 0 and not (solid and bonp) then
		mobj2.momx = mobj2.momx - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -distx)/nerf)
		mobj2.momy = mobj2.momy - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -disty)/nerf)
	end

	-- Do the bump fx when we've CONFIRMED we can bump.
	S_StartSound(mobj1, sfx_s3k49)

	local fx = P_SpawnMobj(mobj1.x/2 + mobj2.x/2, mobj1.y/2 + mobj2.y/2, mobj1.z/2 + mobj2.z/2, MT_BUMP)

	if mobj1.eflags & MFE_VERTICALFLIP then
		fx.eflags = $ | MFE_VERTICALFLIP
	else
		fx.eflags = $ & ~MFE_VERTICALFLIP
	end

	fx.scale = mobj1.scale

	-- Because this is done during collision now, rmomx and rmomy need to be recalculated
	-- so that friction doesn't immediately decide to stop the player if they're at a standstill
	-- Also set justbumped here
	if p1 then
		p1.rmomx = mobj1.momx - p1.cmomx
		p1.rmomy = mobj1.momy - p1.cmomy
		p1.phs_justbumped = bumptime
		if p1.kartstuff[k_spinouttimer] then
			p1.kartstuff[k_wipeoutslow] = wipeoutslowtime+1
			p1.kartstuff[k_spinouttimer] = max(wipeoutslowtime+1, p1.kartstuff[k_spinouttimer])
		end
		if p1.phs_driftsave then
			-- prevent wrongdrifts
			p1.kartstuff[k_driftend] = 0
			p1.kartstuff[k_driftboost] = 0
			p1.kartstuff[k_drift] = p1.phs_driftsave
		end
	end

	if p2 then
		p2.rmomx = mobj2.momx - p2.cmomx
		p2.rmomy = mobj2.momy - p2.cmomy
		p2.phs_justbumped = bumptime
		if p2.kartstuff[k_spinouttimer] then
			p2.kartstuff[k_wipeoutslow] = wipeoutslowtime+1
			p2.kartstuff[k_spinouttimer] = max(wipeoutslowtime+1, p2.kartstuff[k_spinouttimer])
		end

		if p2.phs_driftsave then
			-- prevent wrongdrifts
			p2.kartstuff[k_driftend] = 0
			p2.kartstuff[k_driftboost] = 0
			p2.kartstuff[k_drift] = p2.phs_driftsave
		end
	end
end

local bumpcodeenabled = false

local function KartRiderBumpCode(mo, tm)
	if driftmodData.isOn and bumpcodeenabled
	-- see if it went over / under
	and not ((tm.z > mo.z + tm.height) or (tm.z + tm.height < mo.z))
	and (tm.type == MT_SMK_ICEBLOCK
	or tm.type == MT_PLAYER
	or tm.type == MT_FALLINGROCK
	or tm.type == MT_BLUEROBRA_HEAD
	or tm.type == MT_BLUEROBRA_JOINT
	or tm.type == MT_SMK_PIPE
	or ((tm.type == MT_SMK_THWOMP) and not (mo.z < tm.z and tm.momz < 0))
	or ((tm.flags & MF_SOLID) and (P_IsObjectOnGround(tm) and mo.momz < 0)))
	and P_IsObjectOnGround(mo) and P_IsObjectOnGround(tm) then -- this should only run if both players are on the ground
		KP_KartBouncing(mo, tm, false, true, 12)
	end
end

local function hook()
	addHook("MobjCollide", KartRiderBumpCode, MT_PLAYER)
	addHook("MobjMoveCollide", KartRiderBumpCode, MT_PLAYER)

	addHook("MobjThinker", function(mo)
		if driftmodData.isOn and bumpcodeenabled then
			local p = mo.player
			if p.phs_hititem then
				p.phs_hititem = $ - 1
			else
				p.kartstuff[k_justbumped] = bumptime -- constantly reset justbumped
			end

			if p.phs_justbumped then
				p.phs_justbumped = $ - 1
			else
				p.phs_driftsave = p.kartstuff[k_drift] -- only store if we haven't bumped again
			end

			if not P_IsObjectOnGround(mo) then -- do leeway if you're in midair
				p.phs_justbumped = TICRATE*3/4
			end
		end
	end, MT_PLAYER)

	addHook("ShouldSpin", function(p, i)
		-- G: no more hidden balance changes
		if driftmodData.isOn and bumpcodeenabled and i and i.valid
		and (i.type == MT_ORBINAUT_SHIELD
		or i.type == MT_JAWZ_SHIELD
		or i.type == MT_ORBINAUT
		or i.type == MT_JAWZ
		or i.type == MT_JAWZ_DUD) then
			p.kartstuff[k_justbumped] = 0 -- hit an item
			p.phs_hititem = bumptime
		end
	end)
end

local cv_bumpcode = CV_RegisterVar({
	name = "dfm_bumpcode",
	flags = CV_NETVAR|CV_CALL,
	defaultvalue = "Off",
	possiblevalue = CV_OnOff,
	func = function(cv)
		bumpcodeenabled = cv.value == 1
		if hook and cv.value then
			hook()
			hook = nil
		end
	end
})
